/**
 * pio.c
 *      Another synthetic benchmarking tools for parallel I/O
 *      with FGR support.
 *
 * @author: fwang2@gmail.com
 *
 *
 */
#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif

#define _GNU_SOURCE
#ifndef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#endif

#include <limits.h>         /* for PATH_MAX */
#include <float.h>          /* for DBL_MAX */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <inttypes.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdbool.h>
#include <unistd.h>
#include <getopt.h>
#include <libgen.h>         /* for dirname() */
#include <inttypes.h>
#include <stddef.h>
#include <mpi.h>

#ifdef HAVE_LUSTRE_2
#include <lustre/lustreapi.h>
#include <lustre/lustre_user.h>
#endif

#ifdef HAVE_LUSTRE_18
#include <lustre/liblustreapi.h>
#include <lustre/lustre_user.h>
#endif

#if defined(HAVE_LUSTRE_18) || defined(HAVE_LUSTRE_2)
#define HAVE_LUSTRE
#endif

#ifndef LOV_MAX_STRIPE_COUNT
#define LOV_MAX_STRIPE_COUNT 160
#endif

#ifdef HAVE_PMI_CNOS
void PMI_CNOS_Get_nid(int, int *);
#endif

#include "fgr.h"

typedef enum {
    PIO_LOG_FATAL = 0,
    PIO_LOG_ERR   = 1,
    PIO_LOG_WARN  = 2,
    PIO_LOG_INFO  = 3,
    PIO_LOG_DEBUG = 4,
} PIO_log_t;

typedef enum {
    PIO_DEFAULT = 0,
    PIO_PLACED  = 1,
} PIO_placement_t;

typedef struct {
    int         fd;
    bool        verbose;
    bool        debug;
    bool        exclude_meta;
    bool        donormal;
    bool        dofgr;
    int         iter;
    uint64_t    blocksize;
    uint64_t    xfersize;
    char        outdir[PATH_MAX];
    char        outfile[PATH_MAX];
    long        timestamp;
    int         lustre_stripe_size;
    int         lustre_stripe_count;
    int         lustre_stripe_offset;
    fs_t        fs;
    PIO_placement_t   mode;
} PIO_params_t;

typedef struct {
    time_t  time_started;
    time_t  time_ended;
    double  wtime_started;
    double  wtime_ended;
} PIO_stats_t;

int PIO_rank;
int PIO_size;
int PIO_nid;
uint32_t * PIO_all_nids = NULL;
uint32_t * PIO_all_osts = NULL;
int * PIO_all_ranks = NULL;

double * PIO_write_runs = NULL;
double * PIO_read_runs = NULL;
double * PIO_timing = NULL;

char * PIO_data = NULL;

PIO_params_t PIO_opts;
PIO_stats_t PIO_stats;
PIO_log_t PIO_debug_level = PIO_LOG_INFO;

void PIO_exit(int code) {
    MPI_Finalize();
    exit(code);
}

void PIO_abort(int code) {
    MPI_Abort(MPI_COMM_WORLD, code);
    exit(code);
}

#define PIO_FATAL(...) do {                     \
    if (PIO_LOG_ERR <= PIO_debug_level) {       \
        fprintf(stderr, "PIO-ERR: ");           \
        fprintf(stderr, __VA_ARGS__);           \
        fprintf(stderr, "\n");                  \
        PIO_exit(EXIT_FAILURE);                 \
    }                                           \
} while (0)

#define PIO_ERR(...) do {                       \
    if (PIO_LOG_ERR <= PIO_debug_level) {       \
        fprintf(stderr, "PIO-ERR: ");           \
        fprintf(stderr, __VA_ARGS__);           \
        fprintf(stderr, "\n");                  \
    }                                           \
} while (0)


#define PIO_WARN(...) do {                      \
    if (PIO_LOG_WARN <= PIO_debug_level) {      \
        fprintf(stderr, "PIO-WARN: ");          \
        fprintf(stderr, __VA_ARGS__);           \
        fprintf(stderr, "\n");                  \
    }                                           \
} while (0)

#define PIO_DEBUG(...) do {                     \
    if (PIO_LOG_DEBUG <= PIO_debug_level) {     \
        fprintf(stderr, "PIO-DEBUG: ");         \
        fprintf(stderr, __VA_ARGS__);           \
        fprintf(stderr, "\n");                  \
    }                                           \
} while (0)


#define PIO_INFO(...) do {                      \
    if (PIO_LOG_INFO <= PIO_debug_level) {      \
        fprintf(stdout, __VA_ARGS__);           \
        fprintf(stdout, "\n");                  \
    }                                           \
} while (0)



static inline void
mpi_err_check(int rc, const char * msg) {
    if (rc != MPI_SUCCESS) {
        PIO_FATAL("%s", msg);
    }
}

static const char* get_placement_mode(PIO_placement_t mode) {
    switch(mode)
    {
        case PIO_DEFAULT:
            return "default";
        case PIO_PLACED:
            return "placed";
    }
    return "unknown";
}


static double max_array(double a[], size_t num) {
    double max = -DBL_MAX;
    for (int i = 0; i < num; i++) {
        if (a[i] > max)
            max = a[i];
    }
    return max;
}


static double min_array(double a[], size_t num) {
    double min = DBL_MAX;
    for (int i = 0; i < num; i++) {
        if (a[i] < min)
            min = a[i];
    }
    return min;
}

static double avg_array(double a[], size_t num) {
    double avg = 0.0;
    for (int i = 0; i < num; i++) {
        avg += a[i];
    }
    return avg/num;
}

/* the number conversion utils if from "fileutils" common lib */

static double format_mb(double input) {
    double mb = 1024.0*1024.0;
    while (input/mb > 1.0) {
        input /= mb;
    }
    return input;
}

static void format_1024(
  double input,
  const char** units_list,
  int units_len,
  double* val,
  const char* units[])
{
    /* divide input by 1024 until it falls to less than 1024,
     * increment units each time */
    int idx = 0;
    while(input / 1024.0 > 1.0) {
        input /= 1024.0;
        idx++;
        if(idx == (units_len - 1)) {
            /* we've gone as high as we can go */
            break;
        }
    }

    /* set output paramaters */
    *val   = input;
    *units = units_list[idx];

    return;
}



static const char* units_bytes[] = {"B", "KB", "MB", "GB", "TB", "PB", "EB"};
#define NUM_UNITS_BYTES (7)
/* given a number of bytes, return value converted to returned units */
void format_bytes(uint64_t bytes, double* val, const char** units) {
    double bytes_d = (double) bytes;
    format_1024(bytes_d, units_bytes, NUM_UNITS_BYTES, val, units);
    return;

}

#define NUM_UNITS_BW (6)
static const char* units_bw[] = {"B/s", "KB/s", "MB/s", "GB/s", "TB/s", "PB/s"};

/* given a bandwidth in bytes/sec, return value converted to returned units */
void format_bw(double bw, double* val, const char** units)
{
    format_1024(bw, units_bw, NUM_UNITS_BW, val, units);
    return;
}

static unsigned long long kilo  =                1024ULL;
static unsigned long long mega  =             1048576ULL;
static unsigned long long giga  =          1073741824ULL;
static unsigned long long tera  =       1099511627776ULL;
static unsigned long long peta  =    1125899906842624ULL;
static unsigned long long exa   = 1152921504606846976ULL;


/* abtoull ==> ASCII bytes to unsigned long long
 * Converts string like "10mb" to unsigned long long integer value
 * of 10*1024*1024.  Input string should have leading number followed
 * by optional units.  The leading number can be a floating point
 * value (read by strtod).  The trailing units consist of one or two
 * letters which should be attached to the number with no space
 * in between.  The units may be upper or lower case, and the second
 * letter if it exists, must be 'b' or 'B' (short for bytes).
 *
 * Valid units: k,K,m,M,g,G,t,T,p,P,e,E
 *
 * Examples: 2kb, 1.5m, 200GB, 1.4T.
 *
 * Returns true if conversion is successful,
 * and false otherwise.
 *
 * Returns converted value in val parameter.  This
 * parameter is only updated if successful. */
bool abtoull(const char* str, uint64_t * val) {
    /* check that we have a string */
    if (str == NULL) {
        return false;
    }

    /* check that we have a value to write to */
    if (val == NULL) {
        return false;
    }

    /* pull the floating point portion of our byte string off */
    errno = 0;
    char* next = NULL;
    double num = strtod(str, &next);
    if (errno != 0) {
        /* conversion failed */
        return false;
    }
    if (str == next) {
        /* no conversion performed */
        return false;
    }

    /* now extract any units, e.g. KB MB GB, etc */
    unsigned long long units = 1;
    if (*next != '\0') {
        switch(*next) {
        case 'k':
        case 'K':
            units = kilo;
            break;
        case 'm':
        case 'M':
            units = mega;
            break;
        case 'g':
        case 'G':
            units = giga;
            break;
        case 't':
        case 'T':
            units = tera;
            break;
        case 'p':
        case 'P':
            units = peta;
            break;
        case 'e':
        case 'E':
            units = exa;
            break;
        default:
            /* unknown units symbol */
            return false;
        }

        next++;

        /* handle optional b or B character, e.g. in 10KB vs 10K */
        if (*next == 'b' || *next == 'B') {
            next++;
        }

        /* check that we've hit the end of the string */
        if (*next != 0) {
            return false;
        }
    }

    /* check that we got a positive value */
    if (num < 0) {
        return false;
    }

    /* TODO: double check this overflow calculation */
    /* multiply by our units and check for overflow */
    double units_d = (double) units;
    double val_d = num * units_d;
    double max_d = (double) ULLONG_MAX;
    if (val_d > max_d) {
        /* overflow */
        return false;
    }

    /* set return value */
    *val = (uint64_t) val_d;

    return true;

}

static char* fs2str(fs_t fs) {
    if (fs == ATLAS1)
        return "atlas1";
    else if (fs == ATLAS2)
        return "atlas2";
    else if (fs == ATLAS)
        return "atlas";

    return NULL;
}

void record(int idx, double bw) {
    if ( !PIO_write_runs )
        PIO_write_runs = malloc(PIO_opts.iter * sizeof(double));
    PIO_write_runs[idx] = bw;
}

void create_and_fill_buffer(PIO_params_t *pp, uint64_t offset) {
    if ( !PIO_data )
        PIO_data = malloc(pp->xfersize);

    uint64_t *buf = (uint64_t *) PIO_data;
    uint64_t hi = ((uint64_t) PIO_rank) << 32;
    uint64_t lo = (uint64_t) pp->timestamp;
    for (int i = 0; i < pp->xfersize / sizeof(uint64_t); i++) {
        if ( (i%2) == 0) {
            /* even contains MPI rank and time in seconds */
            buf[i] = hi | lo;
        } else {
            /* odd contains offset */
            buf[i] = offset + (i * sizeof(uint64_t));
        }
    }
}

void create_and_fill_buffer2(PIO_params_t *pp, int value) {
    if ( !PIO_data )
        PIO_data = malloc(PIO_opts.xfersize);
    int * buf = (int *) PIO_data;
    for (int i = 0; i < pp->xfersize / sizeof(int); i++) {
        buf[i] = value;
    }
}

void lustre_open_ioctl(PIO_params_t *pp) {
#ifdef USE_LUSTRE_OPEN_IOCTL
    struct lov_user_md opts = { 0 };
    opts.lmm_magic = LOV_USER_MAGIC;
    opts.lmm_stripe_size = pp->lustre_stripe_size;
    opts.lmm_stripe_offset = pp->lustre_stripe_offset;
    opts.lmm_stripe_count = pp->lustre_stripe_count;
    int fd = open(pp->outfile, O_CREAT | O_EXCL | O_WRONLY | O_LOV_DELAY_CREATE, 0644);
    if (ioctl(fd, LL_IOC_LOV_SETSTRIPE, &opts)) {
        PIO_FATAL("ioctl for '%s': %s", pp->outfile, strerror(errno));
    }
#endif
}

static inline int maxint(int a, int b) {
  return a > b ? a : b;
}

#if 0
static void *alloc_lum() {
  int v1, v3;
  v1 = sizeof(struct lov_user_md_v1) + \
      LOV_MAX_STRIPE_COUNT * sizeof(struct lov_user_ost_data_v1);
  v3 = sizeof(struct lov_user_md_v3) + \
      LOV_MAX_STRIPE_COUNT * sizeof(struct lov_user_ost_data_v1);
  return malloc(maxint(v1, v3));
}
#endif

void
posix_create(PIO_params_t * pp) {

    int fd, rc;

#ifdef HAVE_LUSTRE
    rc = llapi_file_create(pp->outfile, \
            pp->lustre_stripe_size, pp->lustre_stripe_offset, pp->lustre_stripe_count, 0);
    if (rc) {
        PIO_FATAL("llapi_file_create() failed: %s", strerror(-rc));
    }
    if (PIO_rank == 0) {
        PIO_DEBUG("Set stripe count: %d", pp->lustre_stripe_count);
    }

    /* int fd = open(pp->outfile, O_CREAT | O_WRONLY | O_LOV_DELAY_CREATE, 0644); */
    fd = open(pp->outfile, O_WRONLY, 0644);

#else
    fd = open(pp->outfile, O_CREAT | O_TRUNC |O_WRONLY, 0644);
#endif

    if (fd == -1) {
        PIO_FATAL("Failed to open %s", pp->outfile);
    }
    pp -> fd = fd;

}


void
posix_remove(PIO_params_t *pp) {

    int rc1 = access(pp->outfile, F_OK | W_OK);
    if (rc1 == 0) {
        int rc2 = remove(pp->outfile);
        if (rc2) {
            PIO_FATAL("Failed to remove %s: %s", pp->outfile, strerror(errno));
        }
    } else {
        if (errno == EACCES || errno == ETXTBSY) {
            PIO_FATAL("Failed to access existing file %s:", pp->outfile);
        }
    }
}

bool
posix_write(PIO_params_t *pp) {

    off_t len = 0;
    off_t left = pp->blocksize;
    int nr = 0;
    while ( left > 0 ) {
        if (left >= pp->xfersize) {
            len = pp->xfersize;
        } else {
            len = left;
        }

        nr = write(pp->fd, PIO_data, len);
        if (nr == -1) {
            /* error */
            PIO_FATAL("write error: %s", strerror(errno));
        } else if (nr != len) {
            PIO_FATAL("Partial write exception: request [%llu], return [%llu]", \
                    (unsigned long long) len, (unsigned long long) nr);
        }
        left = left - len;

    }

    return true;
}

void PIO_timestamp(PIO_params_t *pp) {
    if (PIO_rank == 0) {
        if ((pp->timestamp = time(NULL)) == -1 ) {
            PIO_FATAL("Can't get current time");
        }
        MPI_Bcast(&pp->timestamp, 1, MPI_LONG, 0, MPI_COMM_WORLD);
    }

}


static void write_debug_file(const char * suffix) {
    char debugfile[PATH_MAX];
    char *jobid = getenv("PBS_JOBID");
    char *jobname = getenv("PBS_JOBNAME");
    if ( jobid == NULL || jobname == NULL)
        return;
    sprintf(debugfile, "%s-o%s.%s", jobname, jobid, suffix);
    int fd = open(debugfile, O_CREAT | O_WRONLY | O_APPEND, 0644);
    char * buf = malloc(PIO_size * 8);
    sprintf(buf, "%d: ", PIO_size);
    for (int i = 0; i < PIO_size; i++) {
        if (strcmp(suffix, "nids") == 0)
            sprintf(buf + strlen(buf), "%d ", PIO_all_nids[i]);
        else
            sprintf(buf + strlen(buf), "%d ", PIO_all_osts[i]);
    }
    sprintf(buf + strlen(buf), "\n");
    int nr = write(fd, buf, strlen(buf));
    if (nr != strlen(buf)) {
        PIO_WARN("debug file write() expected: %zu, return: %d, reason: %s",
                strlen(buf), nr, strerror(errno));
    }
    close(fd);
    if (buf) {
        free(buf);
    }

}

void PIO_prep_fgr(PIO_params_t *pp) {
    struct stat st;
    bool rc;
    char * config_path = getenv("PIO_CONFIG_PATH");
    if (config_path != NULL) {
        rc = fgr_init(config_path);
    } else {
        stat("/lustre/atlas2/test/fwang2/bpio", &st);
        if (S_ISDIR(st.st_mode)) {
            rc = fgr_init("/lustre/atlas2/test/fwang2/bpio");
        } else {
            rc = fgr_init(".");
        }
    }

    if (! rc && PIO_rank == 0) {
        PIO_FATAL("fgr_init() failed");
    }

    fgr_nid2ost(PIO_all_nids, PIO_all_osts, PIO_size, pp->fs);


    if (PIO_LOG_DEBUG == PIO_debug_level && PIO_rank == 0) {
        fgr_rtr_debug();
        fgr_lnet_debug();
        fgr_oss_debug();
        fgr_ost_debug();
    }

    fgr_finalize();

    if (PIO_rank == 0) {
        write_debug_file("osts");
    }

    pp->lustre_stripe_offset = PIO_all_osts[PIO_rank];
    pp->lustre_stripe_count = 1;

}

/*
 * FIXME: multiple runs with same MPI size, the records
 * can't tell them apart.
 */
static void
write_timing_file(PIO_params_t* pp, int iter) {
    char tfile[512];
    char buf[4096];

    char* jobid = getenv("PBS_JOBID");
    char* jobname = getenv("PBS_JOBNAME");
    if (jobid && jobname) {
        sprintf(tfile, "%s-o%s.timing", jobname, jobid);
    } else {
        sprintf(tfile, "pio-timing.dat");
    }

    int fd = open(tfile, O_CREAT | O_WRONLY | O_APPEND, 0644);
    if (fd == -1) {
        PIO_FATAL("file creation: %s", strerror(errno));
    }

    for (int rank = 0; rank < PIO_size; rank ++) {
        /*
         * mode, iter, size
         */
        sprintf(buf, "%s, %d, %d, %d", get_placement_mode(pp->mode), iter, PIO_size, rank);
#ifdef HAVE_LUSTRE
        sprintf(buf + strlen(buf), ", %d, %d, %d",
                pp->lustre_stripe_size,
                pp->lustre_stripe_count,
                pp->mode == PIO_DEFAULT? pp->lustre_stripe_offset: PIO_all_osts[rank]);
#endif

        sprintf(buf + strlen(buf), ", %" PRIu64 ", %" PRIu64,
                pp->blocksize, pp->xfersize);

        sprintf(buf + strlen(buf), ", %f", PIO_timing[rank]);

        sprintf(buf + strlen(buf), "\n");

        int nr = write(fd, buf, strlen(buf));
        if (nr != strlen(buf)) {
            PIO_WARN("write() expected: %zu, return: %d, reason: %s", strlen(buf), nr, strerror(errno));
        }
    }
    close(fd);
}

void PIO_write_run(PIO_params_t* pp, int iter) {
    int rc;

    if (PIO_timing == NULL && PIO_rank == 0) {
        PIO_timing = malloc(PIO_size * sizeof(double));
    }

    PIO_timestamp(pp);
    posix_remove(pp);
    create_and_fill_buffer(pp, rand());
    if (pp->exclude_meta) {
        posix_create(pp);
        PIO_stats.wtime_started = MPI_Wtime();
    } else {
        PIO_stats.wtime_started = MPI_Wtime();
        posix_create(pp);
    }
    posix_write(pp);


    fsync(pp->fd);

    if (pp->exclude_meta) {
        PIO_stats.wtime_ended = MPI_Wtime();
        close(pp->fd);
    } else {
        close(pp->fd);
        PIO_stats.wtime_ended = MPI_Wtime();
    }

/*
    struct lov_user_md * lum_file = alloc_lum();
    if (lum_file == NULL && PIO_rank == 0) {
      PIO_FATAL("alloc_lum() failed: %s", strerror(errno));
    }
    rc = llapi_file_get_stripe(pp->outfile, lum_file);

    PIO_DEBUG("pp->outfile = %s", pp->outfile);

    if (rc && PIO_rank == 0) {
        rc = errno;
        PIO_FATAL("llapi_file_get_stripe() failed: %s", strerror(rc));
    }

    PIO_DEBUG("rank_id= %d, stripe_count= %d, stripe_offset= %d",
        PIO_rank,
        lum_file->lmm_stripe_count,
        lum_file->lmm_stripe_offset);

    if (lum_file != NULL)
      free(lum_file);
*/

    double delta = PIO_stats.wtime_ended - PIO_stats.wtime_started;
    double delta_max;
    rc = MPI_Reduce(&delta, &delta_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    mpi_err_check(rc, "MPI_Reduce for delta max failed");
    MPI_Gather(&delta, 1, MPI_DOUBLE, PIO_timing, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    mpi_err_check(rc, "MPI_Gather for delta failed");

    if (PIO_rank == 0) {
        double bw = (PIO_size * PIO_opts.blocksize) / delta_max;
        record(iter, bw);
        PIO_INFO("Iteration: %d, Rate: %.3lf MiB/s", iter, format_mb(bw));
        write_timing_file(pp, iter);
    }
}

void PIO_write_prelogue(const char * keyword, int argc, char ** argv) {
    if (PIO_rank == 0) {
        char buf[4096];
        sprintf(buf, "\nBench Mode: %s", keyword);
        PIO_INFO("%s", buf);

        sprintf(buf, "Command line: ");
        for (int i = 1; i < argc; i++)
            sprintf(buf + strlen(buf), "%s ", argv[i]);
        PIO_INFO("%s", buf);
    }
}

void PIO_write_epilogue() {
    if (PIO_rank == 0) {
        char buf[4096];
        double piomax = max_array(PIO_write_runs, PIO_opts.iter);
        sprintf(buf, "PIO.Write.Max = %.3lf MiB/s, ", format_mb(piomax));

        double piomin = min_array(PIO_write_runs, PIO_opts.iter);
        sprintf(buf + strlen(buf), "PIO.Write.Min = %.3lf MiB/s, ", format_mb(piomin));

        double pioavg = avg_array(PIO_write_runs, PIO_opts.iter);
        sprintf(buf + strlen(buf), "PIO.Write.Avg = %.3lf MiB/s\n", format_mb(pioavg));

        PIO_INFO("%s", buf);
    }
}

void PIO_version() {
#ifdef HAVE_CONFIG_H
    fprintf(stdout, "%s\n", BPIO_VERSION);
#endif
    exit(0);
}

void PIO_usage(char ** argv) {
    char * opts[] = {
            "Options:",
            "-a N   stripe_offset -- Lustre stripe offset (default -1)",
            "-b N   blocksize -- bytes to write per task (eg. 100m, 8g)",
            "-d     debug -- set debug level",
            "-c N   stripe_count -- Lustre stripe count (default 4)",
            "-d S   debug options (TBI)",
            "-e     exclude metadata operation",
            "-f S   filesystem -- one of atlas1, atlas2 or atlas",
            "-h     help -- print usage help",
            "-i N   iter -- run with specific iterations number (default is 3)",
            "-n     normal -- run with default placement",
            "-o S   outdir -- output directory, must pre-exists",
            "-p     placement -- enable FGR-based placement",
            "-s N   stripe_size -- Lustre stripe size (default 1m)",
            "-t N   xfersize -- bytes for each write (default is 1m)",
            "-v     version - print build version",
            " ",
            "Note: S is string, N is number, all units are 1024-based",
            "PIO performs fsync() by default at the end.",
            " ",
            ""
    };
    fprintf(stdout, "Usage: %s [OPTIONS]\n\n", *argv);
    for (int i = 0; strlen(opts[i]) > 0; i++) {
        fprintf(stdout, "%s\n", opts[i]);
    }
}

void PIO_refresh() {
    if (!PIO_read_runs)
        free(PIO_read_runs);
    if (!PIO_write_runs)
        free(PIO_write_runs);
    if (!PIO_data)
        free(PIO_data);
}

void PIO_finalize() {
    PIO_refresh();
    if (!PIO_all_nids)
        free(PIO_all_nids);

    if (!PIO_all_osts)
        free(PIO_all_osts);
}

void create_nids_buffer() {
    PIO_all_nids = malloc(PIO_size * sizeof(int));
    if (PIO_all_nids == NULL)
        PIO_FATAL("create_nids_buffer() memory alloc failed");

    PIO_all_osts = malloc(PIO_size * sizeof(int));
    if (PIO_all_osts == NULL)
        PIO_FATAL("create_ost_buffer() memory alloc failed");
}

int main(int argc, char** argv) {
    int c;
    int option_index = 0;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &PIO_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &PIO_size);
#ifdef HAVE_PMI_CNOS
    PMI_CNOS_Get_nid(PIO_rank, &PIO_nid);
    PIO_DEBUG("Rank = %d, nid = %d", PIO_rank, PIO_nid);
    create_nids_buffer();
    MPI_Allgather(&PIO_nid, 1, MPI_INT, PIO_all_nids, 1 , MPI_INT, MPI_COMM_WORLD);

    if (PIO_rank == 0) {
        write_debug_file("nids");
    }
#endif
    /* seed the random */
    srand(time(NULL));

    /* set default */
    PIO_opts.iter = 3;
    PIO_opts.donormal = false;
    PIO_opts.exclude_meta = false;
    PIO_opts.dofgr = false;
    PIO_opts.lustre_stripe_offset = -1;
    PIO_opts.lustre_stripe_count = 4;

    abtoull("1mb", &PIO_opts.xfersize);
    abtoull("8gb", &PIO_opts.blocksize);

    if (PIO_opts.xfersize > INT_MAX ) {
        PIO_FATAL("PIO_opts.xfersize: %" PRIu64 " > %d", PIO_opts.xfersize, INT_MAX);
    }
    PIO_opts.lustre_stripe_size = (int) PIO_opts.xfersize;

    DIR *dir = NULL;
    static struct option long_options[] = {
            {"stripe_offset", required_argument, 0, 'a'},
            {"blocksize"    , required_argument, 0, 'b'},
            {"stripe_count" , required_argument, 0, 'c'},
            {"debug"        , required_argument, 0, 'd'},
            {"exclude"      , no_argument,       0, 'e'},
            {"filesystem"   , required_argument, 0, 'f'},
            {"help"         , no_argument,       0, 'h'},
            {"iter"         , required_argument, 0, 'i'},
            {"normal"       , no_argument,       0, 'n'},
            {"outdir"       , required_argument, 0, 'o'},
            {"placement"    , no_argument,       0, 'p'},
            {"stripe_size"  , required_argument, 0, 's'},
            {"version"      , no_argument,       0, 'v'},
            {"xfersize"     , required_argument, 0, 't'},
            {NULL}
    };

    /* FIXME: more error checking, replace atoi() with something better */

    while ((c = getopt_long(argc, argv, "a:b:c:d:ef:hi:no:ps:t:v", \
            long_options, &option_index)) != -1) {
        switch (c) {
        case 'a':
            PIO_opts.lustre_stripe_offset = atoi(optarg);
            break;
        case 'b':
            if (!abtoull(optarg, &PIO_opts.blocksize)) {
                if (PIO_rank == 0)
                    PIO_FATAL("Failed to convert -b: %s", optarg);
            }
            break;
        case 'c':
            PIO_opts.lustre_stripe_count = atoi(optarg);
            break;
        case 'd':
            PIO_debug_level = PIO_LOG_DEBUG;
            break;
        case 'e':
            PIO_opts.exclude_meta = true;
            break;
        case 'f':
            if (! strcmp(optarg, "atlas1")) {
                PIO_opts.fs = ATLAS1;
            } else if (!strcmp(optarg, "atlas2")) {
                PIO_opts.fs = ATLAS2;
            } else if (!strcmp(optarg, "atlas")) {
                PIO_opts.fs = ATLAS;
            } else {
                if (PIO_rank == 0)
                    PIO_FATAL("Failed to convert -f: %s", optarg);
            }
            break;
        case 'n':
            PIO_opts.donormal = true;
            break;
        case 'p':
            PIO_opts.dofgr = true;
            break;
        case 's':
            if (!abtoull(optarg, (uint64_t *) &PIO_opts.lustre_stripe_size)) {
                if (PIO_rank == 0)
                    PIO_FATAL("Failed to convert -s: %s", optarg);
            }
            break;
        case 't':
            if (!abtoull(optarg, &PIO_opts.xfersize)) {
                if (PIO_rank == 0)
                    PIO_FATAL("Failed to convert %s", optarg);
            }
            break;
        case 'i':
            PIO_opts.iter = atoi(optarg);
            if (PIO_opts.iter == 0) {
                if (PIO_rank == 0)
                    PIO_FATAL("Failed to convert %s", optarg);
            }
            break;
        case 'o':
            dir = opendir(optarg);
            if (dir) {
                size_t len = strlen(optarg);
                if ((len > 0) && (optarg[len-1] == '/'))
                    optarg[len-1] = '\0';
                strcpy(PIO_opts.outdir, optarg);
            } else {
                if (PIO_rank == 0)
                    PIO_FATAL("Directory: %s doesn't exist, must quit!", optarg);
            }
            break;
        case 'h':
            PIO_usage(argv);
            PIO_exit(EXIT_FAILURE);
            break;
        case 'v':
            /* PIO_debug_level = PIO_LOG_DEBUG; */
            PIO_version();
            break;
        case '?':
        default:
            if (PIO_rank == 0) {
                if (isprint(optopt)) {
                    PIO_usage(argv);
                    PIO_FATAL("Unknown option `-%c'.\n", optopt);
                }
            }
            break;
        } /* switch */
    } /* while */

    if (dir == NULL && PIO_rank == 0) {
        PIO_ERR("Must provide an valid output path");
        PIO_usage(argv);
        PIO_exit(EXIT_FAILURE);
    }

    if ( !PIO_opts.donormal && !PIO_opts.dofgr) {
        if (PIO_rank == 0)
            PIO_FATAL("placement mode: -n or -p missing");
    }


    if (PIO_rank == 0) {
 #ifdef HAVE_LUSTRE
        PIO_INFO("Have Lustre:          Yes");
        PIO_INFO("Lustre stripe size:   %d", PIO_opts.lustre_stripe_size);
        PIO_INFO("Lustre stripe count:  %d", PIO_opts.lustre_stripe_count);
        PIO_INFO("Lustre stripe offset: %d", PIO_opts.lustre_stripe_offset);
#else
        PIO_INFO("Have Lustre:          No");
#endif

#ifdef HAVE_PMI_CNOS
        PIO_INFO("Have PMI CNOS:        Yes");
#else
        PIO_INFO("Have PMI CNOS:        No");
#endif
        PIO_INFO("Filesystem:           %s", fs2str(PIO_opts.fs));
        PIO_INFO("Transfer size:        %" PRIu64, PIO_opts.xfersize);
        PIO_INFO("Block size:           %" PRIu64, PIO_opts.blocksize);
        PIO_INFO("Output path:          [%s]", PIO_opts.outdir);
    }

    sprintf(PIO_opts.outfile, "%s/pio.%08d", PIO_opts.outdir, PIO_rank);


    if (PIO_opts.donormal) {
        PIO_opts.mode = PIO_DEFAULT;
        PIO_write_prelogue("DEFAULT", argc, argv);
        for (int i = 0; i < PIO_opts.iter; i++) {
            PIO_write_run(&PIO_opts, i);
        }
        PIO_write_epilogue();
    }

    if (PIO_opts.dofgr) {
        PIO_opts.mode = PIO_PLACED;
        PIO_write_prelogue("PLACED", argc, argv);
        PIO_refresh();
        PIO_prep_fgr(&PIO_opts);
        for (int i = 0; i < PIO_opts.iter; i++) {
            PIO_write_run(&PIO_opts, i);
        }

        PIO_write_epilogue();
    }

    PIO_finalize();
    PIO_exit(EXIT_SUCCESS);

}
